#!/usr/bin/perl

# City of Heroes installer/patcher/launcher
# 
# Copyright (c) 2011 Mark Wagner
# No Rights Reserved

# Version history:
#	0.5: First version.
#	0.5.1: Fixed some residual references to /tmp
#	0.6: First version tested against an actual update.  Fixed path-quoting and return value issues.
#	0.6.1: Fixed a bug where the script would give a bogus "unable to create directory" error.
#	0.6.2: Changed from "system" to "exec" for launching CoH.
#	       Added "--silent" and "--silentlaunch" options.
#	0.7: Added support for Additions.xml.
#	     Added support for launching CoH in languages other than English (--en, --de, --fr options)
#	0.8: Added support for repair patches (--repair option)
#	     Made Configure() report all missing programs at once, rather than one per run.
#	     Fixed detection of unrar.
#	     Set unrar to automatically over-write existing files.
#	0.8.1: Fixed retrieval of Additions.xml
#	0.8.2: Added "--patchonly" option.
#	       Added support for FileDelete operations.
#	       Added the ability to resume a partially-applied patch.
#	0.8.3: Added improved checks for unknown patch operations.
#	       Suppressed redundant file verification by the CoH client.
#          Fixed a crash when there's only one FileAdd, FileDelete, or FilePatch operation.
#	0.8.4: Replaced shellouts to 'pwd' with the Cwd module.
#          Added support for Digest::MD5 and libwww-perl.
#          Added preliminary support for Crossover Games based on third-party reports (untested).

# Must be run from the City of Heroes directory.  Running from an empty directory
# will install a new copy of City of Heroes, which is probably not what you want.

use warnings;
use strict;
use XML::Simple;
use File::Copy;
use File::Path;
use Cwd;

my $version = "0.8.4";

my $source;
my $keep_patches = 0;
my $silent = 0;
my $silent_launch = 0;
my $language = 'en';
my $region = 'us';
my $repair = 0;
my $patchonly = 0;

my $get_program;
my $xdelta_program;
my $unrar_program;
my $md5_program;

my @args;

my $parser = XML::Simple->new(ForceArray => ['FileAdd', 'FileDelete', 'FilePatch']);

# Verify that all needed programs are available, and set up basic patching environment
sub Configure
{
	my $missing_programs = 0;

	# Either wget or curl
	if(system('wget --version >/dev/null 2>&1') == 0)
	{
		print "Using wget for file transfers.\n" if(!$silent);
		$get_program = 'wget -c -N ';
	}
	elsif(system("curl --version >/dev/null 2>&1") == 0)
	{
		print "Using curl for file transfers.\n" if(!$silent);
		$get_program = 'curl -R -O ';
		# NOTE: Transfer resuming ("-C -") is not used because cURL returns an error code
		# if the complete file has already been downloaded.
	}
	elsif(eval{ require LWP::UserAgent })
	{
		print "Using LWP::UserAgent for file transfers.\n" if(!$silent);
		$get_program = '_INTERNAL';
	}
	else
	{
		print STDERR "This script requires wget, curl, or libwww-perl to be installed.\n";
		$missing_programs = 1;
	}
	
	# Xdelta
	my $ver_statement = `xdelta --version 2>&1`;
	$ver_statement = '' if(!defined($ver_statement));
	if($ver_statement =~ /version 1.1/)
	{
		print "Using xdelta for patching.\n" if(!$silent);
		$xdelta_program = 'xdelta';
	}
	elsif($ver_statement =~ /version 3/)
	{
		print STDERR "This script requires xdelta 1.  Xdelta 3 is not compatible with version 1 patches.\n";
		$missing_programs = 1;
	}
	else
	{
		print STDERR "This script requires xdelta 1 to be installed.\n";
		$missing_programs = 1;
	}
	
	# Unrar
	my $unrar_return = system('unrar >/dev/null 2>&1') >> 8;
	if($unrar_return >= 0 && $unrar_return < 127 )
	{
		print "Using unrar for decompression.\n" if(!$silent);
		$unrar_program = 'unrar x -o+ ';
	}
	else
	{
		print STDERR "This script requires unrar to be installed.\n";
		$missing_programs = 1;
	}
	
	#MD5
	if(system('md5sum /dev/null >/dev/null 2>&1') == 0)
	{
		print "Using md5sum for MD5 checksums.\n" if(!$silent);
		$md5_program = 'md5sum';
	}
	elsif(eval{ require Digest::MD5 })
	{
		print "Using Digest::MD5 for MD5 checksums.\n" if(!$silent);
		$md5_program = '_INTERNAL';
	}
	else
	{
		print "md5sum not found.  Patch checksums will not be verified.\n";
	}

	exit(1) if($missing_programs);

	# Set up patch temp directory
	mkdir '.patches' if(! -e '.patches');
	die "Unable to create directory '.patches'\n" if(! -e '.patches');
}

# Parse the command line
sub ParseOptions
{
	my @args;
	my $renderthread_seen = 0;
	
	foreach my $arg (@_)
	{
		if($arg eq '-test' or $arg eq '--test' or $arg eq '-cohtest' or $arg eq 'cohtest')
		{
			$source = 'cohpts';
			$region = 'us';
		}
		elsif($arg eq '-beta' or $arg eq '--beta' or $arg eq '-cohbeta' or $arg eq 'cohbeta')
		{
			$source = 'cohbeta';
			$region = 'us';
		}
		elsif($arg eq '-live' or $arg eq '--live' or $arg eq '-coh' or $arg eq 'coh')
		{
			$source = 'cohlive';
			$region = 'us';
		}
		elsif($arg eq '-eulive' or $arg eq '--eulive' or $arg eq '-eucoh' or $arg eq 'eucoh')
		{
			$source = 'cohlive';
			$region = 'eu';
		}
		elsif($arg eq '-project')
		{
			# Discard
		}
		elsif($arg eq '-keep' or $arg eq '--keep')
		{
			$keep_patches = 1;
		}
		elsif($arg eq '--silentlaunch')
		{
			$silent_launch = 1;
		}
		elsif($arg eq '--silent')
		{
			$silent = 1;
			$silent_launch = 1;
		}
		elsif($arg eq '-renderthread')
		{
			$renderthread_seen = 1;
			push @args, $arg;
		}
		elsif($arg eq '--en')
		{
			$language = 'en';
		}
		elsif($arg eq '--de')
		{
			$language = 'de';
		}
		elsif($arg eq '--fr')
		{
			$language = 'fr';
		}
		elsif($arg eq '--repair')
		{
			$repair = 1;
		}
		elsif($arg eq '--patchonly' or $arg eq '--updateonly')
		{
			$patchonly = 1;
		}
		elsif($arg eq '--help' or $arg eq '-help')
		{
			print "Usage: cohlaunch.pl [--patchonly] [--keep] [--repair] [--silentlaunch] [--silent] [(--en|--de|--fr)] (-live|-eulive|-test|-beta) [City of Heroes options].\n";
			print "\n";
			print "-live, -eulive, -test, -beta: Update for the US, European, test,\n";
			print "\tor beta clients.\n";
			print "\tChanging an installation from US/EU to test, or US/EU to beta is\n";
			print "\tsupported, while changing from test to beta may or may not work.\n";
			print "\tChanging from beta or test to US/EU is not supported and will\n";
			print "\tprobably not work.\n";
			print "--en, --de, --fr: Select the language to use.\n";
			print "--keep: Keep downloaded patches.  This may speed up future patching, or\n";
			print "\tit may simply waste large amounts of disk space.  It's your choice.\n";
			print "\tThe launcher will delete patches that fail checksum verification\n";
			print "\tregardless of this setting.\n";
			print "--repair: Verify file checksums and re-download files if needed.\n";
			print "\tFiles are transferred uncompressed rather than as compressed patches,\n";
			print "\tso it is not recommended that you repair an out-of-date installation\n";
			print "\tunless updating fails.\n";
			print "--silentlaunch: Suppress all console output from City of Heroes.\n";
			print "\tThis may improve stability or give a slight increase in framerate.\n";
			print "\tFor most people, I expect it will do nothing.\n";
			print "--silent: Suppress all output except error messages.\n";
			print "--patchonly: Only patch the City of Heroes client, do not launch it.\n";
			print "\n";
			print "Passing the City of Heroes option '-renderthread 0' may be required, as\n";
			print "multithreaded rendering tends not to work on Linux.\n";
			print "\n";
			print "Run this script from an existing installation directory, or\n";
			print "from an empty directory to install a new copy of City of Heroes.\n";
			print "\n";
			print "This script has not been extensively tested; use at your own risk.\n";
			print "\n";
			print "cohlaunch.pl version $version\n";
			print "Copyright (c) 2011-2012 Mark Wagner.  No Rights Reserved.\n";
			
			exit;
		}
		else
		{
			push @args, $arg;
		}
	}
	$source = 'cohlive' if(!defined($source));
	print "Warning: City of Heroes may require the '-renderthread 0' option to function properly\n" if(!$renderthread_seen and (!$silent and !$patchonly));
	return @args;
}

sub GetFile
{
	my $file = shift;
	my $result;
	
	# http://cohbeta.patcher.ncsoft.com/CityOfHeroes/Main/Manifests/PatchListManifest.xml
	# http://cohpts.patcher.ncsoft.com/CityOfHeroes/Main/Manifests/PatchListManifest.xml
	# http://cohlive.patcher.ncsoft.com/CityOfHeroes/Main/Manifests/PatchListManifest.xml
	if($get_program eq '_INTERNAL')
	{
		my ($target_file) = $file =~ /.*\/(.*)/;
		my $ua = LWP::UserAgent->new(env_proxy => 1);
		$ua->show_progress(1) if(!$silent);
		my $response = $ua->get("http://$source.patcher.ncsoft.com/CityOfHeroes/Main/$file", ':content_file' => $target_file);
		if($response->is_success())
		{
			$result = 0;
		}
		else
		{
			print $response->status_line(), "\n";
			$result = 1;
		}
	}
	else
	{
		if($silent)
		{
			$result = system("$get_program http://$source.patcher.ncsoft.com/CityOfHeroes/Main/$file >/dev/null 2>&1");
		}
		else
		{
			$result = system("$get_program http://$source.patcher.ncsoft.com/CityOfHeroes/Main/$file");
		}
	}
	die "Unable to fetch file.\n" if($result != 0);
}

sub UnpackFile
{
	my $file = shift;

	if($silent)
	{
		system("$unrar_program '$file' >/dev/null 2>&1") && die "Unable to decompress $file\n"; # unrar returns 0 for success
	}
	else
	{
		system("$unrar_program '$file'") && die "Unable to decompress $file\n";	# unrar returns 0 for success
	}
}

sub GetManifest
{
	my $base = cwd();
	chdir(".patches") || die "Unable to change working directory to '.patches'.\n";
	unlink('PatchListManifest.xml') if(-e 'PatchListManifest.xml');
	die "Unable to remove existing .patches/PatchListManifest.xml\n" if(-e 'PatchListManifest.xml');
	GetFile('Manifests/PatchListManifest.xml');
	chdir($base) || die "Unable to change to installation base directory.\n";
}

# Parse a version string into a tuple
sub ParseVersion
{
	my $ver = shift;
	my ($major, $minor, $patch, $p2) = $ver =~ /(\d*)(?:\.(\d*)(?:\.(\d*)(?:\.(\d*)|)|)|)/;
	$minor = 0 if(!defined($minor));
	$patch = 0 if(!defined($patch));
	$p2 = 0 if(!defined($p2));
	return ($major, $minor, $patch, $p2);
}

# Is the first version newer than the second?
sub VersionGreater
{
	my $v1 = shift;
	my $v2 = shift;
	my ($v1_major, $v1_minor, $v1_patch, $v1_p2) = ParseVersion($v1);
	my ($v2_major, $v2_minor, $v2_patch, $v2_p2) = ParseVersion($v2);
	return 1 if($v1_major > $v2_major);
	return 0 if($v1_major < $v2_major);
	return 1 if($v1_minor > $v2_minor);
	return 0 if($v1_minor < $v2_minor);
	return 1 if($v1_patch > $v2_patch);
	return 0 if($v1_patch < $v2_patch);
	return 1 if($v1_p2 > $v2_p2);
	return 0; # Less than or equal to
}

# Are two versions equal?
sub VersionEqual
{
	my $v1 = shift;
	my $v2 = shift;
	my ($v1_major, $v1_minor, $v1_patch, $v1_p2) = ParseVersion($v1);
	my ($v2_major, $v2_minor, $v2_patch, $v2_p2) = ParseVersion($v2);
	return ($v1_major == $v2_major and $v1_minor == $v2_minor and $v1_patch == $v2_patch and $v1_p2 == $v2_p2);
}

sub GetInstalledVersion
{
	my $current_ver;

	if(-r "version.ini")
	{
		open VERFILE, '<', "version.ini";
		while($_ = <VERFILE>)
		{
			if($_ =~ /^Version=/)
			{
				($current_ver) = $_ =~ /^Version\s*=\s*([0-9.]*)/;
				last;
			}
		}
		close VERFILE;
	}
	
	if(!defined($current_ver))
	{
		$current_ver = '0.0.0.0';
	}
	return $current_ver;
}

sub GetLatestVersion
{
	my $target_ver = '0.0.0.0';
	my $parsed_xml = $parser->XMLin('.patches/PatchListManifest.xml');
	
	foreach my $patch (@{$parsed_xml->{PatchListManifest}->{Patches}->{Patch}})
	{
		$target_ver = $patch->{TargetVersion} if($patch->{Type} eq 'VersionTargeted' and VersionGreater($patch->{TargetVersion}, $target_ver));
	}
	return $target_ver;
}

sub IsLatest
{
	my $current_ver = GetInstalledVersion();
	my $target_ver = GetLatestVersion();
	if(VersionEqual($current_ver, '0.0.0.0'))
	{
		print "Unable to determine current version of City of Heroes; installing from scratch\n" if(!$silent);
		return 0;
	}
	elsif(VersionEqual($current_ver, $target_ver))
	{
		print "City of Heroes is up to date\n" if(!$silent);
		return 1;
	}
	elsif(VersionGreater($target_ver, $current_ver))
	{
		print "City of Heroes needs updating\n" if(!$silent);
		return 0;
	}
	else
	{
		die "Installed version is newer than the version on the server.\nYou may be trying to convert a beta or test install to a live install, which is not supported.\n";
	}
}

sub VerifyMD5
{
	my $file = shift;
	my $MD5 = shift;
	my $file_MD5;
	
	if(!defined($md5_program))
	{
		return 1;	# If we can't check, return success
	}
	elsif($md5_program eq '_INTERNAL')
	{
		open INFILE, '<', $file;
		my $ctx = Digest::MD5->new();
		$ctx->addfile(*INFILE);
		$file_MD5 = $ctx->hexdigest();
		close INFILE;
	}
	else
	{
		$file_MD5 = `$md5_program '$file'`;
		$file_MD5 =~ s/([0-9a-fA-F]*).*/${1}/;
		chomp $file_MD5;
	}
#	print "Expected: " . lc($MD5) . " Got: " . lc($file_MD5) . "x\n";
	return (lc($MD5) eq lc($file_MD5));
}

sub ApplyPatch
{
	my $operations_performed = 0;

	my $file = shift;
	my $name = shift;
	my $MD5 = shift;
	my $COH_path = cwd();
	chdir('.patches') || die "Unable to change working directory to '.patches'.\n";
	
	# Fetch the patch
	GetFile("Patches/$file");
	if(!VerifyMD5($file, $MD5))
	{
		unlink($file);
		die "Patch $file failed MD5 checksum verify.\n";
	}

	# Unpack the patch
	UnpackFile($file);
	
	chdir($name) || die "Unable to change to patch directory $name\n";
	my $parsed_xml = $parser->XMLin('PatchManifest.xml');

	# Sanity checks -- make sure the patch isn't asking us to do something we don't know how to.
	die "Unknown patch format.\n" if(!exists($parsed_xml->{Version}));
	print "Warning: patch version $parsed_xml->{Version} is not a known version.  Patching may fail.\n" if(!VersionEqual($parsed_xml->{Version}, '1.1.0.0'));
	die "This is a repair patch.\n" if(exists($parsed_xml->{IsRepair}) and lc($parsed_xml->{IsRepair} eq 'true'));
	die "This patch has unknown steps.\n" if(grep {$_ ne 'ApplyFileGroup'} keys(%{$parsed_xml->{PatchApplication}->{Steps}}));
	die "This patch is not targeting the install directory.\n" if($parsed_xml->{PatchApplication}->{settings}->{property}->{PatchTargetDirectory}->{content} ne '${_ProductInstallPath}');
	die "This patch has unknown application requirements.\n" if(scalar(keys(%{$parsed_xml->{PatchApplication}->{ApplicationRequirements}})));
	
	my @unknown_operations = grep {$_ !~ /(FileDelete|FileAdd|FilePatch|Name)/} keys(%{$parsed_xml->{PatchFileGroup}});
	die "This patch has unknown file operations: " . join(", ", @unknown_operations) . ".\n" if(scalar(@unknown_operations));

	# Apply and verify the patch
	my $group_name = $parsed_xml->{PatchFileGroup}->{Name};
	foreach my $patch (@{$parsed_xml->{PatchFileGroup}->{FileDelete}})
	{
		# Files to delete
		my $path = $patch->{Path};
		$path =~ s/\\/\//g;
		if(-e "${COH_path}/${path}" and VerifyMD5("${COH_path}/${path}", $patch->{Hash}))
		{
			unlink("${COH_path}/${path}");
		}
		else
		{
			print STDERR "Unable to delete file ${path}.\n";
		}
		
		$operations_performed++;
	}
	foreach my $patch (@{$parsed_xml->{PatchFileGroup}->{FileAdd}})
	{
		# Files to copy
		my $path = $patch->{Path};
		$path =~ s/\\/\//g;
		my ($target_dir) = $path =~ /^(.*\/)/;
		$target_dir = "" if(!defined($target_dir));
		unlink("${COH_path}/${path}") if(-e "${COH_path}/${path}");
		if(! -d "${COH_path}/${target_dir}")
		{
			File::Path::make_path("${COH_path}/${target_dir}") || die "Unable to make directory ${COH_path}${target_dir}\n";
		}
		File::Copy::move("${group_name}/${path}", "${COH_path}/${path}");
		if(!VerifyMD5("${COH_path}/${path}", $patch->{Hash}))
		{
			die "Copying file $path failed.\n";
		}
		$operations_performed++;
	}
	foreach my $patch (@{$parsed_xml->{PatchFileGroup}->{FilePatch}})
	{
		# Binary patches to apply
		
		my $path = $patch->{Path};
		$path =~ s/\\/\//g;
		if($silent)
		{
			system("$xdelta_program patch '${group_name}/${path}.X-D-E-L-T-A' '${COH_path}/${path}' '${COH_path}/${path}.patched' >/dev/null 2>&1");
		}
		else
		{
			system("$xdelta_program patch '${group_name}/${path}.X-D-E-L-T-A' '${COH_path}/${path}' '${COH_path}/${path}.patched'");
		}
		if(!VerifyMD5("${COH_path}/${path}.patched", $patch->{HashPostPatch}))
		{
			if(!VerifyMD5("${COH_path}/${path}", $patch->{HashPostPatch}))
			{
				die "Patching file $path failed.\n";
			}
			else
			{
				unlink("${COH_path}/${path}.patched");
				print "File ${path} is already patched; skipping\n" if(!$silent);
			}
		}
		else
		{
			unlink("${COH_path}/${path}");
			File::Copy::move("${COH_path}/${path}.patched", "${COH_path}/${path}");
		}
		$operations_performed++;
	}
	
	die "This patch did not do anything.\n" if(!$operations_performed);

	# Clean up
	chdir($COH_path);
	File::Path::remove_tree(".patches/$name", {safe => 1});
	unlink(".patches/$file") if(!$keep_patches);
}

sub ApplyRepairPatch
{
	my $file = shift;
	my $name = shift;
	my $MD5 = shift;
	my $COH_path = cwd();
	chdir('.patches') || die "Unable to change working directory to '.patches'.\n";

	# Fetch the patch
	GetFile("Patches/$file");
	if(!VerifyMD5($file, $MD5))
	{
		unlink($file);
		die "Patch $file failed MD5 checksum verify.\n";
	}

	# Unpack the patch
	UnpackFile($file);
	
	chdir($name) || die "Unable to change to patch directory $name\n";
	my $parsed_xml = $parser->XMLin('PatchManifest.xml');

	die "Unknown patch format.\n" if(!exists($parsed_xml->{Version}));
	print "Warning: patch version $parsed_xml->{Version} is not a known version.  Patching may fail.\n" if(!VersionEqual($parsed_xml->{Version}, '1.1.0.0'));
	die "This is an upgrade patch.\n" if(lc($parsed_xml->{IsRepair} ne 'true'));
	die "This patch has unknown steps.\n" if(grep {$_ ne 'ApplyFileGroup'} keys(%{$parsed_xml->{PatchApplication}->{Steps}}));
	die "This patch is not targeting the install directory.\n" if($parsed_xml->{PatchApplication}->{settings}->{property}->{PatchTargetDirectory}->{content} ne '${_ProductInstallPath}');
	die "This patch has unknown application requirements.\n" if(scalar(keys(%{$parsed_xml->{PatchApplication}->{ApplicationRequirements}})));

	# Apply and verify the patch
	foreach my $patch (@{$parsed_xml->{PatchFileGroup}->{FileUnchanged}})
	{
		my $path = $patch->{Path};
		$path =~ s/\\/\//g;
		my $fetch_needed = 0;
		
		print "Verifying $path..." if(!$silent);
		
		$fetch_needed = 1 if(! -e "${COH_path}/${path}");
		$fetch_needed = 1 if(!VerifyMD5("${COH_path}/${path}", $patch->{Hash}));
		
		if($fetch_needed)
		{
			print "Failed\n" if(!$silent);
			
			my ($filename) = $path =~ /([^\/]*)$/;
			# Note: The "?h=" parameter does not appear to be essential; rather, it appears to be used for
			# cache-managing purposes, to ensure that if we get the file from a cache (say, the Akami CDN that
			# NCSoft uses), we get the right version.
			GetFile("Repair/ProductFiles/${path}?h=$patch->{Hash}");
			File::Copy::move("${filename}?h=$patch->{Hash}", $filename) if(-e "${filename}?h=$patch->{Hash}");
			if(VerifyMD5($filename, $patch->{Hash}))
			{
				my ($target_dir) = $path =~ /^(.*\/)/;
				$target_dir = "" if(!defined($target_dir));
				unlink("${COH_path}/${path}") if(-e "${COH_path}/${path}");
				if(! -d "${COH_path}/${target_dir}")
				{
					File::Path::make_path("${COH_path}/${target_dir}") || die "Unable to make directory ${COH_path}${target_dir}\n";
				}
				File::Copy::move($filename, "${COH_path}/${path}");
				if(!VerifyMD5("${COH_path}/${path}", $patch->{Hash}))
				{
					die "Copying file $path failed.\n";
				}
			}
			else
			{
				die "Fetching file $path failed.\n";
			}
		}
		else
		{
			print "OK\n" if(!$silent);
		}
	}

	# Clean up
	chdir($COH_path);
	File::Path::remove_tree(".patches/$name", {safe => 1});
	unlink(".patches/$file") if(!$keep_patches);
}

sub RepairCoh
{
	my $current_ver = GetInstalledVersion();
	my $parsed_xml = $parser->XMLin('.patches/PatchListManifest.xml');
	my $patched = 0;
	
	# Find a repair patch for a version at least as new as what's installed
	foreach my $patch (@{$parsed_xml->{PatchListManifest}->{Patches}->{Patch}})
	{
		if($patch->{Type} eq 'Repair' and (VersionGreater($patch->{TargetVersion}, $current_ver) || VersionEqual($patch->{TargetVersion}, $current_ver)))
		{
			print "Warning: Repair patch is for a newer version of City of Heroes than is currently installed.  This may fail.\n" if(VersionGreater($patch->{TargetVersion}, $current_ver));
			ApplyRepairPatch($patch->{RelativePath}, $patch->{PatchName}, $patch->{FileHash});
			last;
		}
	}
}

sub PatchCoh
{
	my $current_ver = GetInstalledVersion();
	my $latest_ver = GetLatestVersion();
	my $parsed_xml = $parser->XMLin('.patches/PatchListManifest.xml');
	my $patched = 0;
	
	while(!VersionEqual($current_ver, $latest_ver))
	{
		# Find a patch that will take us from our current version to a newer version.
		# We assume that "FromVersion" in the patch manifest is unique, and that the
		# (FromVersion, ToVersion) pairs form a directed acyclic graph with its root 
		# at the current version.  If this is not true, we might find ourselves in
		# an infinite loop or stuck unable to patch an older version.
		foreach my $patch (@{$parsed_xml->{PatchListManifest}->{Patches}->{Patch}})
		{
			if($patch->{Type} eq 'VersionTargeted' and VersionEqual($patch->{FromVersion}, $current_ver))
			{
				print "Patching from $current_ver to $patch->{TargetVersion} using $patch->{PatchName}\n" if(!$silent);
				ApplyPatch($patch->{RelativePath}, $patch->{PatchName}, $patch->{FileHash});
				$patched = 1;
				$current_ver = $patch->{TargetVersion};
				last;
			}
		}
		if(!$patched)
		{
			die "Unable to find an upgrade path from version $current_ver.  To force a reinstall, delete the file 'version.ini'.\n";
		}
		$patched = 0;
	}
}

sub GetAdditionalArgs
{
	my @additional_args;
	my $additional_args;
	
	my $base = cwd();
	chdir(".patches") || die "Unable to change working directory to '.patches'.\n";
	unlink 'Additions.xml' if(-e 'Additions.xml');
	GetFile('Additions.xml');
	chdir($base) || die "Unable to change to installation base directory.\n";	
	
	if(-e '.patches/Additions.xml')
	{
		my $additions = $parser->XMLin('.patches/Additions.xml');
		if($source eq 'cohlive')
		{
			if($region eq 'us')
			{
				foreach my $r (@{$additions->{Game_Arguments}->{CityOfHeroes}->{Value}})
				{
					$additional_args = $r->{content} if(lc($r->{Regions}) eq $region);
				}
			}
			else
			{
				foreach my $r (@{$additions->{Game_Arguments}->{CityOfHeroesEU}->{Value}})
				{
					if(lc($r->{Regions}) eq $region)
					{
						$additional_args = $r->{content};
						foreach my $lang (@{$r->{Value}})
						{
							$additional_args = $lang->{content} if(lc($lang->{Languages}) eq $language);
						}
					}
				}
			}
		}
		elsif($source eq 'cohbeta')
		{
			foreach my $r (@{$additions->{Game_Arguments}->{CoHBeta}->{Value}})
			{
				$additional_args = $r->{content} if(lc($r->{Regions}) eq $region);
			}
		}
		elsif($source eq 'cohpts')
		{
			foreach my $r (@{$additions->{Game_Arguments}->{CoHTest}->{Value}})
			{
				$additional_args = $r->{content} if(lc($r->{Regions}) eq $region);
			}
		}
		else
		{
			die "Unknown download source: $source\n";
		}
		
		chomp $additional_args;
		$additional_args =~ s/^\s*//;
		$additional_args =~ s/\s*$//;
		@additional_args = split / /, $additional_args;
	}
	else
	{
		print "Warning: Unable to retrieve additional command-line parameters from the server.  Guessing at reasonable defaults.";
		if($source eq 'cohlive')
		{
			if($region eq 'eu')
			{
				push @additional_args, ('-project', 'eucoh', '-launcher', '-auth', '64.25.36.4');
			}
			else
			{
				push @additional_args, ('-project', 'coh', '-launcher', '-auth', '64.25.36.4');
			}
		}
		elsif($source eq 'cohbeta')
		{
			push @additional_args, ('-project', 'cohbeta', '-launcher', '-auth', '64.25.36.19');
		}
		elsif($source eq 'cohpts')
		{
			push @additional_args, ('-project', 'cohtest', '-launcher', '-auth', '64.25.36.21');
		}
		else
		{
			die "Unknown download source: $source\n";
		}
	}

	# We do our own file verification.  No need to waste the user's time by
	# having CoH do it as well.
	push @additional_args, '-noverify';
	return @additional_args;
}

sub LaunchCoh
{
	my @args = @_;
	my $command;
	if(-x "/opt/cxgames/bin/wine")
	{
		$command = '/opt/cxgames/bin/wine --cx-app cityofheroes.exe ' . join(" ", @args);
	}
	elsif(-x "$ENV{HOME}/cxgames/bin/wine")
	{
		$command = '~/cxgames/bin/wine --cx-app cityofheroes.exe ' . join(" ", @args);
	}
	else
	{
		$command = "wine cityofheroes.exe " . join(" ", @args);
	}
	$command .= " >/dev/null 2>&1" if($silent_launch);
	exec($command);
}

@args = ParseOptions(@ARGV);
Configure();
GetManifest();
if($repair)
{
	RepairCoh();
}
else
{
	if(!IsLatest())
	{
		PatchCoh();
	}
}
if(!$patchonly)
{
	# Get Additions.xml and parse for extra command-line arguments
	my @server_args = GetAdditionalArgs();
	LaunchCoh(@server_args, @args);
}

